import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import minimize


def load_mpg_as_dataframe():
    """Load 'auto-mpg.csv' and return the disp and hp columns as numpy arrays."""
    df = pd.read_csv('auto-mpg.csv', skipinitialspace=True,
                     names=['mpg', 'cyl', 'disp', 'hp', 'wgt', 'acc', 'year', 'orig', 'name'])
    return df


def load_data_for_univariate_regression():
    """Load 'auto-mpg.csv' and return the disp and hp columns as numpy arrays."""
    df = load_mpg_as_dataframe()
    X = df.disp.values[:, np.newaxis]
    y = df.hp.values
    return X, y


def load_data_for_multivariate_regression():
    """Load 'auto-mpg.csv' and return the disp and hp columns as numpy arrays."""
    df = load_mpg_as_dataframe()
    X = df.xs(['mpg', 'cyl', 'disp', 'wgt', 'acc'], axis=1).values
    y = df.hp.values
    return X, y


def plot_2d_data(X, y):
    """Plot the data in 2D space, assuming X is 1-dimensional."""
    plt.figure()
    plt.scatter(X, y)
    plt.xlabel('disp', fontsize=16)
    plt.ylabel('hp', fontsize=16)


def plot_predictions(w, c='r', label=None):
    """Plot predictions of the linear model with parameters w."""
    # Assuming there exists a plot of the data,
    # read the axes limits:
    ax = plt.axis()
    # Build the test data as xmin and xmax:
    Xtst = np.array(ax[0:2])[:, np.newaxis]
    # Find the predictions
    ytst = pred_regr_lin(w, Xtst)
    # Plot them
    plt.plot(Xtst, ytst, lw=3, c=c, label=label)
    # Set axes limits as they were before
    plt.axis(ax)
    plt.legend(loc='upper left')


def homogenize(X):
    """Return X with a prepended column of ones."""
    # FIXME Task 2: homogenize X
    raise NotImplementedError


def pred_regr_lin(w, X):
    """Predict the values of y for inputs X, given the linear model parameters w."""
    # FIXME Task 3: predict y_hat
    raise NotImplementedError


def compute_err_mse(y, yhat):
    """Return the mean squared error given the true values of y and predictions yhat."""
    # FIXME Task 4: compute MSE
    raise NotImplementedError


def compute_cost_regr_lin(w, X, y):
    """Return the cost of lin. reg. and its gradient at point w, given data X and y."""
    yhat = pred_regr_lin(w, X)
    err = compute_err_mse(y, yhat)
    return err


def fit_regr_lin_by_minimization(X, y):
    """Return the parameters of linear model trained by MSE minimization."""
    # FIXME Task 6: minimize with scipy
    raise NotImplementedError
    # Define f(w) somehow, see the exercise description doc.
    # w_init = np.array([0,0])
    # res = minimize( ... )
    # w = res.x
    # return w


def fit_regr_lin_by_normal_equation(X, y):
    """Return the parameters of linear model using normal equation."""
    # FIXME Task 7: normal equation
    raise NotImplementedError
